﻿var FACEBOOK_GET_TOKEN_REGEX = /<auth_createToken_response.*>(.*)<\/auth_createToken_response>/;
var FACEBOOK_GET_SESSION_KEY = /<session_key>(.*)<\/session_key>/;
var FACEBOOK_GET_UID = /<uid>(.*)<\/uid>/g;
var FACEBOOK_GET_AID = /<aid>(.*)<\/aid>/g;
var FACEBOOK_GET_PHOTO = /<photo>([\s\S]*)<\/photo>/g;
var FACEBOOK_GET_USER = /<user>([\s\S]*)<\/user>/g;

var FACEBOOK_GET_UID_NOT_GLOBAL = /<uid>(.*)<\/uid>/;
var FACEBOOK_GET_SRC = /<src_big>(.*)<\/src_big>/;
var FACEBOOK_GET_OWNER = /<owner>(.*)<\/owner>/;
var FACEBOOK_GET_NAME = /<name>(.*)<\/name>/;
var FACEBOOK_GET_LINK = /<link>(.*)<\/link>/;
var FACEBOOK_GET_SECRET = /<secret>(.*)<\/secret>/;
var FACEBOOK_URL_PREFIX = "https://api.facebook.com/login.php";
var FACEBOOK_URL_LOGOUT = "https://www.facebook.com/logout.php";
var FACEBOOK_URL_API_KEY = "api_key";
var FACEBOOK_URL_AUTH_TOKEN = "auth_token";
var FACEBOOK_URL_OPTIONS = "popup=true";

var FACEBOOK_API_CREATE_TOKEN = "auth.createToken";
var FACEBOOK_API_GET_SESSION = "auth.getSession";
var FACEBOOK_API_GET_FRIENDS = "friends.get";
var FACEBOOK_API_GET_PHOTO_ALBUMS = "photos.getAlbums";
var FACEBOOK_API_GET_PHOTOS = "photos.get";
var FACEBOOK_API_GET_USERSINFO = "users.getInfo";
var FACEBOOK_API_PARAMETER_AUTH_TOKEN = "auth_token";
var FACEBOOK_API_PARAMETER_USER_ID = "uid";
var FACEBOOK_API_PARAMETER_ALBUM_ID = "aid";
var FACEBOOK_API_PARAMETER_USER_IDS = "uids";
var FACEBOOK_API_PARAMETER_USER_FIELDS = "fields";
var FACEBOOK_API_ARGUMENT_USER_FIELDS = "name";

var FACEBOOK_REQUEST_CONTENT_TYPE = "application/x-www-form-urlencoded;charset=UTF-8;";
var FACEBOOK_LOGOUT_TEXT = "confirm=1";

//  - check if user is already logged in by doing http request first
//  - return structures of friends, albums, etc
//  - redesign using interfaces?
//  - ASPECT RATIO
//  - get feedback from interns at google

// Ideas for this gadget
//   - Animations to make pictures come on screen. Perhaps fade small then get larger while moving?
//   - When offline, go through images that would be stored offline
//   - Maybe scroll comments
//   - Display caption
//   - When hoving over the picture, maybe highlight it to show its clickable, thatd be cool
//   - When hovering, show the PLAY PAUSE BACK FORWARD buttons!!
//   - Make video and collect feedback from interns?


// The Facebook class provides a wrapper to the Facebook API documented at
// http://developers.facebook.com/index.php.
// the onTokenExpiry_ function will be called if the token ever becomes invalid
// The client will need to handle this situation.
function Facebook(apiKey_, originalSecret_, onTokenExpiry_) {
  // Whenever the parent does a Facebook request, this is done asynchronously.
  // Therefore, a request variable is required to keep track of the call in case
  // of requiring to stop it using the .stop() method.
  var currentRequest_ = null;

  // Since in a callback the exact error is not returned, the getLastError and
  // getLastErrorCode functions will return the detailed information. The
  // details are stored in these variables when the error occurs. 'null'
  // represents no error. Error code of -1 is an unknown error. Error of ""
  // also represents an unknown error.
  var lastErrorCode_ = null; 
  var lastError_ = null; 

  // This information is returned from Facebook after doing requests. All are
  // part of the authentication sequence. For more information on their
  // requirement, please see the Facebook API.
  var authToken_ = null;          // Authentication Token
  var sessionKey_ = null;         // Session Key
  var uid_ = null;                // Logged in user ID
  var secret_ = null;             // Current secret (changes once authenticated)

  // Set the session variables so a login isn't required
  function setSession(session, uid, secret) {
    sessionKey_ = session;
    uid_ = uid;
    secret_ = secret;
  }

  // Get the session variables for storing
  function getSession() {
    var session = new Object();

    session.sessionKey = sessionKey_;
    session.uid = uid_;
    session.secret = secret_;

    return session;
  }

  // Log out of facebook
  function logout() {
    logoutRequest = createXhr();

    try {
      logoutRequest.open("POST", FACEBOOK_URL_LOGOUT, true);
      logoutRequest.setRequestHeader("Content-type", FACEBOOK_REQUEST_CONTENT_TYPE);
    } catch (e) {
      return;
    }

    // Send the request
    logoutRequest.onreadystatechange = onLoggedOut;

    try {
      logoutRequest.send(FACEBOOK_LOGOUT_TEXT);
    } catch(e) {
      return;
    }

    debug.trace("logging out!")
  }

  // Called once the logout has completed.
  function onLoggedOut() {
    debug.trace("logged out!");
  }

  // Request an authentication URL from facebook servers
  function getAuthenticationURL(onReceivedURL) {   
    // Reset the secret to the main application secret
    secret_ = originalSecret_;

    // Start the request
    currentRequest_ = createFacebookRequest(FACEBOOK_API_CREATE_TOKEN, 
      function (data) {
        if (checkForRequestError(data, onReceivedURL))
          return;

        // Obtain the token from the response
        authToken_ = getFirstRegex(FACEBOOK_GET_TOKEN_REGEX, data);
        if (authToken_ == null) {
          lastErrorCode_ = -1;
          onReceivedURL(null);
          return;
        }

        // Generate the URL and return it to the parent
        var url = FACEBOOK_URL_PREFIX + "?" + FACEBOOK_URL_API_KEY + "=" + 
          apiKey_ + "&" + FACEBOOK_URL_AUTH_TOKEN + "=" + authToken_ + "&" + 
          FACEBOOK_URL_OPTIONS;

        onReceivedURL(url);
      }
    );

    currentRequest_.send();
  }

  // Attempt to authenticate. This should be called after 1) an auth_token has
  // already been received by using getAuthenticationURL and 2) the user
  // has successfully logged in on the returned URL from #1.
  function authenticate(onAuthenticationResult) {    
    // Verify that a valid auto token exists, if not error. The application
    // will need to run getAuthenticationURL method.
    if (authToken_ == null) {
      // Create an ASYNC failure callback
      setTimeout(function () { onAuthenticationResult(false); }, 1);
      return;
    }
  
    currentRequest_ = createFacebookRequest(FACEBOOK_API_GET_SESSION, 
      function (data) {
        if (checkForRequestError(data, onAuthenticationResult))
          return;

        // Get the session key, uid and NEW secret
        sessionKey_ = getFirstRegex(FACEBOOK_GET_SESSION_KEY, data);
        uid_ = getFirstRegex(FACEBOOK_GET_UID, data);
        secret_ = getFirstRegex(FACEBOOK_GET_SECRET, data);

        if ((sessionKey_ == null) || (uid_ == null) || (secret_ == null)) {
          onAuthenticationResult(false);
          return;
        }

        onAuthenticationResult(true);
      }
    );

    currentRequest_.addParameter(FACEBOOK_API_PARAMETER_AUTH_TOKEN, authToken_);
    currentRequest_.send();
  }

  // Returns a list of friends
  function getFriends(onReceivedFriendsList) {
    // Require that a session key was already obtained
    if (sessionKey_ == null) {
      setTimeout(function () { onReceivedFriendsList(null); }, 1);
      return;
    }

    currentRequest_ = createFacebookRequest(FACEBOOK_API_GET_FRIENDS, 
      function (data) {
        if (checkForRequestError(data, onReceivedFriendsList))
          return;

        var uidResult = null;
        var friends = [];
        while ((uidResult = FACEBOOK_GET_UID.exec(data)) != null) {
          if (uidResult.length >= 2) {
            friends.push(uidResult[1]);
          }
        }

        onReceivedFriendsList(friends);
      }
    );
    currentRequest_.send();
  }

  function getFriendsInfo(friends, onReceivedFriendsInfo) {
    // Require that a session key was already obtained
    if (sessionKey_ == null) {
      setTimeout(function () { onReceivedFriendsInfo(null, friends); }, 1);
      return;
    }

    currentRequest_ = createFacebookRequest(FACEBOOK_API_GET_USERSINFO, 
      function (data) {
        if (checkForRequestError(data, onReceivedFriendsInfo, friends))
          return;

        var userResult = null;
        var friendsInfo = [];
        while ((userResult = FACEBOOK_GET_USER.exec(data)) != null) {
          if (userResult.length >= 2) {
            var indivData = userResult[1];
            var uid = FACEBOOK_GET_UID_NOT_GLOBAL.exec(indivData);
            var name = FACEBOOK_GET_NAME.exec(indivData);
            if (uid && uid.length >= 2 && name && name.length >=2) {
              friendsInfo.push([uid[1], name[1]]);
            } else {
              debug.error("FIELD PARSE ERROR!!!!");
            }
          } else {
            debug.error("USER PARSE ERROR!!!!");
          }
        }

        onReceivedFriendsInfo(friendsInfo, friends);
      }
    );
    currentRequest_.addParameter(FACEBOOK_API_PARAMETER_USER_IDS, friends.join(','));
    currentRequest_.addParameter(FACEBOOK_API_PARAMETER_USER_FIELDS, 
      FACEBOOK_API_ARGUMENT_USER_FIELDS);

    currentRequest_.send();
  }

  // Returns a list of albums
  function getAlbums(uid, onReceivedAlbumsList) {
    // Require that a session key was already obtained
    if (sessionKey_ == null) {
      onReceivedAlbumsList(null, uid);
      return;
    }

    currentRequest_ = createFacebookRequest(FACEBOOK_API_GET_PHOTO_ALBUMS, 
      function (data) {
        if (checkForRequestError(data, onReceivedAlbumsList, uid))
          return;

        currentRequest_ = null;

        var aidResult = null;
        var albums = [];
        
        while ((aidResult = FACEBOOK_GET_AID.exec(data)) != null) {
          if (aidResult.length >= 2) {
            albums.push(aidResult[1]);
          }
        }

        onReceivedAlbumsList(albums, uid);

      }
    );
    currentRequest_.addParameter(FACEBOOK_API_PARAMETER_USER_ID, uid);
    currentRequest_.send();
  }

  // Returns a list of photos in an album
  function getPhotos(aid, onReceivedPhotosList) {
    // Require that a session key was already obtained
    if (sessionKey_ == null) {
      onReceivedPhotosList(null, aid);
      return;
    }

    currentRequest_ = createFacebookRequest(FACEBOOK_API_GET_PHOTOS, 
      function (data) {
        if (checkForRequestError(data, onReceivedPhotosList, aid))
          return;

        var photoResult = null;
        var photos = [];
        
        // Get the url of the photo
        while ((photoResult = FACEBOOK_GET_PHOTO.exec(data)) != null) {
          if (photoResult.length >= 2) {
            var photoData = photoResult[1];

            // Get the link of the photo
            var image = FACEBOOK_GET_SRC.exec(photoData);
            var link = FACEBOOK_GET_LINK.exec(photoData);
            var owner = FACEBOOK_GET_OWNER.exec(photoData);
            if (image && image.length >= 2 && link && link.length >= 2 &&
                owner && owner.length >= 2) {
              
              // Create the photo object
              var photoObj = new Object();
              photoObj.src = image[1];
              photoObj.link = link[1];
              photoObj.owner = owner[1];

              photos.push(photoObj);
            }
          }
        }

        onReceivedPhotosList(photos, aid);
      }
    );

    currentRequest_.addParameter(FACEBOOK_API_PARAMETER_ALBUM_ID, aid);
    currentRequest_.send();
  }

  // Check if there was an error with the given data return from the current
  // Facebook request. If so, set last error and callback parent notifying
  // them of the error.
  function checkForRequestError(data, resultCallback, opt_tag) {
    if (data == null) {
      if (currentRequest_) {
        lastError_ = currentRequest_.getLastError();
        lastErrorCode_ = currentRequest_.getLastErrorCode();
        debug.error("request has error " + lastErrorCode_ + " " + lastError_);
        
        // error 102 is the facebook error for 
        // "Session key invalid or no longer valid." Handle this
        if (lastErrorCode_ == 102) {
          debug.error("Session key is invalid. Tell client via onTokenExpiry_");
          setTimeout(onTokenExpiry_, 1);
          currentRequest_ = null;
          return true;
        }
      }

      // Make an ASYNC call back notifying of an error.
      setTimeout(function () { resultCallback(null, opt_tag); }, 1);
    }
     
    currentRequest_ = null;
    return (data == null);
  }

  // Creates a facebook request. Returns a FacebookRequest object with the new
  // request.
  function createFacebookRequest(method, resultCallback) {
    // Reset the last error
    lastError_ = null;
    lastErrorCode_ = null;

    var request = new FacebookRequest(apiKey_, secret_, sessionKey_, method, 
      resultCallback);

    return request;
  }

  // Returns the last error that occured. If no error occurred, null is 
  // returned. If "" is returned, an unknown error occured.
  function getLastError() {
    if (lastErrorCode_ == -1)
      return "";

    return lastError_;
  }

  // Returns the last error code that occurred. If no error occurred, null is 
  // returned. If -1 is returned, an unknown error occurred.
  function getLastErrorCode() {
    return lastErrorCode_;
  }

  // Executes the given regular expression and returns the first result. If 
  // the regular expression fails, null is returned.
  function getFirstRegex(regex, data) {
    var regexResult = regex.exec(data);
    if ((regexResult == null) || (regexResult.length != 2))
      return null;

    return regexResult[1];
  }

  // Stops all current request
  function stop() {
    if (currentRequest_ != null) {
      currentRequest_.stop();
      currentRequest_ = null;
    }
  }

  this.authenticate = authenticate;
  this.getAlbums = getAlbums;
  this.getAuthenticationURL = getAuthenticationURL;
  this.getFriends = getFriends;
  this.getFriendsInfo = getFriendsInfo;
  this.getLastError = getLastError;
  this.getLastErrorCode = getLastErrorCode;
  this.getPhotos = getPhotos;
  this.getSession = getSession;
  this.setSession = setSession;
  this.logout = logout;
  this.stop = stop;
}
