var FACEBOOK_ERROR_EXISTS = /<error_response/;
var FACEBOOK_ERROR_GET_CODE = /<error_code>(.*)<\/error_code>/;
var FACEBOOK_ERROR_GET_MSG = /<error_msg>(.*)<\/error_msg>/;
var FACEBOOK_ERROR_GET_METHOD = /<arg>[\s\S]*<key>method<\/key>[\s\S]*?<value>(.*)<\/value>/;
var FACEBOOK_REQUEST_POST_URL = "https://api.facebook.com/restserver.php";
var FACEBOOK_API_VERSION = "1.0";

var FACEBOOK_MAX_ERRORS_IN_A_ROW = 5;
/**
 * The minimum retry delay when we receive an error from Facebook.
 * @type Number
 */
var BASE_REQUEST_RETRY_MS = 5000;

/**
 * The multiple representing no backoff.
 * @type Number
 */
var BASE_REQUEST_INTERVAL_MULTIPLE = 1;
/**
 * The maximum backoff multiple.
 * @type Number
 */
var MAX_REQUEST_INTERVAL_MULTIPLE = 60;
/**
 * How much we have backed off.
 * @type Number
 */
var g_requestIntervalMultiple = BASE_REQUEST_INTERVAL_MULTIPLE;
/**
 * Roughly how fast should we be backing off, if necessary.
 * @type Number
 */
var REQUEST_INTERVAL_BACKOFF_FACTOR = 2;

/**
 * Resets the backoff
 */
function resetBackoff() {
  debug.trace("resetting backoff");
  g_requestIntervalMultiple = BASE_REQUEST_INTERVAL_MULTIPLE;
}

/**
 * Decrease the rate of facebook requests, so we don't hose the servers.
 */
function backoff() {
  debug.trace("backing off");
  g_requestIntervalMultiple = Math.min(
    g_requestIntervalMultiple * REQUEST_INTERVAL_BACKOFF_FACTOR,
    MAX_REQUEST_INTERVAL_MULTIPLE);
}

// Globally keeps track of how many requests have been made since Facebook
// requires that each call has a unique and increasing ID.
var facebookRequestCurrentCallID_ = (new Date()).getTime();

// The FacebookRequest class provides a method request data from Facebook
// using a POST HTTP request. receivedResultCallback_ will be called with null
// if there was an unrecoverable Facebook error.
function FacebookRequest(apiKey_, secret_, sessionKey_, method_, receivedResultCallback_) {
  var lastErrorCode_ = null;        // Null represents no error, -1 Unknown
  var lastError_ = null;            // Null represents no error
  var errors_ = 0;                  // Total errors from the same request
  var retryTimer_ = null;           // Timer used to retry when a recoverable error occurs

  // The FacebookRequestParam structure describes one parameter for a request
  // to facebook using the Facebook API
  function FacebookRequestParameter(name_, value_) {
    this.name_ = name_;
    this.value_ = value_;
  }

  var params_ = [];     // Contains an array of FacebookRequestParameter
  var request_ = null;  // The XMLHttpRequest done to the facebook servers
  var stop_ = false;    // Keep track if the current request has been stopped

  addParameter("method", method_);
  addParameter("api_key", apiKey_);
  addParameter("v", FACEBOOK_API_VERSION);
  addParameter("call_id", ++facebookRequestCurrentCallID_);

  if (sessionKey_ != null)
    addParameter("session_key", sessionKey_);

  // Send the request to the Facebook and capture the output. If an error
  // occurs, this will return null.
  function send() {
    // Sort all parameters such that the signature will be correct. Parameter
    // lists will always be very small, so a bubble sort will do.
    for (var x = 0; x < params_.length; x++) {
      for (var y = 0; y < params_.length - 1; y++) {
        if (params_[y].name_ > params_[y + 1].name_) {
          var tmpName = params_[y + 1].name_;
          var tmpValue = params_[y + 1].value_;

          params_[y + 1].name_ = params_[y].name_;
          params_[y + 1].value_ = params_[y].value_;

          params_[y].name_ = tmpName;
          params_[y].value_ = tmpValue;
        }
      }
    }

    // Generate string for signature and POST (they differ in the use of not
    // having a "&" and having one)
    var paramsString = "";
    var sigParamsString = "";
    for (var i = 0; i <= params_.length - 1; i++) {
      if (i > 0)
        paramsString += "&";

      // REMOVE_GADGET_BEGIN
      // TODO: do you really need to uri encode for a POST? when you uri
      // encode commas here, that breaks a request, but leaving a comma as is
      // works fine.
      // TODO: try encodeURI. will not encode commas among other chars.
      // REMOVE_GADGET_END
      var component = encodeURIComponent(params_[i].name_) + "=" + params_[i].value_; //encodeURIComponent(params_[i].value_);
      paramsString += component;
      sigParamsString += component;
    }

    // Calculate signature, NOTE, MD5 routine is required, see
    // http://www.pajhome.org.uk/crypt/md5/ for obtaining this.
    sigParamsString += secret_;
    var signature = hex_md5(sigParamsString);

    // Update the parameters string with the signature appended to the end
    paramsString += "&sig=" + signature;

    // Begin the request to facebook servers.
    request_ = createXhr();

    try {
      request_.open("POST", FACEBOOK_REQUEST_POST_URL, true);
      request_.setRequestHeader("Content-type", FACEBOOK_REQUEST_CONTENT_TYPE);
    } catch (e) {
      lastErrorCode_ = -1;
      receivedResultCallback_(null);
      request_ = null;
      return false;
    }

    // Send the request
    request_.onreadystatechange = onFacebookData;
    stop_ = false;

    try {
      request_.send(paramsString);
    } catch(e) {
      lastErrorCode_ = -1;
      receivedResultCallback_(null);
      request_ = null;
      return false;
    }

    return true;
  }

  // Called by the XMLHttpRequest request_ object when data is incoming
  // from Facebook.
  function onFacebookData() {
    // If the download has been stopped, do not continue
    if (stop_)
      return;

    // If the data is not ready (still downloading), do not continue
    if (request_.readyState != 4)
      return;

    // If there was an http error, do not continue
    if (request_.status != 200) {
      backoff();
      lastErrorCode_ = -1;
      receivedResultCallback_(null);
      request_ = null;
      return;
    }

    // Check if an error was returned from Facebook
    var data = request_.responseText;
    if (!isSuccessful(data)) {
      // Get the error code, message and method
      var errorCode = getFirstRegex(FACEBOOK_ERROR_GET_CODE, data);
      var errorMsg = getFirstRegex(FACEBOOK_ERROR_GET_MSG, data);
      var errorMethod = getFirstRegex(FACEBOOK_ERROR_GET_METHOD, data);

      // Check if this was a malformed error message
      if ((errorCode == null) || (errorMsg == null) || (errorMethod == null)) {
        lastErrorCode_ = -1;
      } else {
        // Return a string structured error message
        lastError_ = errorCode + " - " + errorMsg + " - " + errorMethod;
        lastErrorCode_ = errorCode;
      }

      // This might be a recoverable error by adding a delay
      errors_++;
      if ((errors_ < FACEBOOK_MAX_ERRORS_IN_A_ROW) &&
          ((errorCode == 1) || (errorCode == 2) || (errorCode == 4))) {

        var timeout; 
        if (g_requestIntervalMultiple == BASE_REQUEST_INTERVAL_MULTIPLE) {
          timeout = BASE_REQUEST_RETRY_MS;
        } else {
          // timeout randomly in the later half of the interval
          var half_interval = BASE_REQUEST_RETRY_MS * 
            g_requestIntervalMultiple / 2;
          timeout = half_interval + Math.random() * half_interval;
        }
          
        debug.trace("Recoverable error (code " + errorCode + ") waiting " + 
          timeout + "ms for retry...");
        retryTimer_ = setTimeout(retrySend, timeout);
        
        backoff();
      } else {
        receivedResultCallback_(null);
      }

      request_ = null;
      return;
    }

    // Return the requested data
    receivedResultCallback_(request_.responseText);
    
    // reset the backoff since we have success
    resetBackoff();
    request_ = null;
  }

  // Called to retry a send after a recoverable error
  function retrySend() {
    retryTimer_ = null;
    send();
  }

  // Checks if the XML data is a Facebook error. If so, the lastError value is
  // set and will return false. Otherwise, this will return true.
  function isSuccessful(data) {
    // If the data was null, this must be a failure
    if (data == null)
      return false;

    // Check for the failure pattern (if it does not exist, return success)
    var errorRegex = FACEBOOK_ERROR_EXISTS.exec(data);
    if ((errorRegex == null) || (errorRegex.length != 1))
      return true;

    return false;
  }

  // Adds a parameter to the facebook request. The name and value are
  // automatically formatted for the POST request.
  function addParameter(name, value) {
    var param = new FacebookRequestParameter(name, value);
    params_.push(param);
  }

  // Stop a request
  function stop() {
    if (request_ == null)
      return;

    if (retryTimer_ != null) {
      clearInterval(retryTimer_);
      retryTimer_ = null;
    }

    stop_ = true;
    request_.abort();
    request_ = null;
  }

  // Returns the last error that occurred. If no error occured, null is
  // returned. If "" is returned, an unknown error occured.
  function getLastError() {
    if (lastErrorCode_ == -1)
      return "";

    return lastError_;
  }

  // Returns the last error code that occurred. If no error occurred, null is
  // returned. If -1 is returned, an unknown error occurred.
  function getLastErrorCode() {
    return lastErrorCode_;
  }

  // Executes the given regular expression and returns the first result. If
  // the regular expression fails, null is returned.
  function getFirstRegex(regex, data) {
    var regexResult = regex.exec(data);
    if ((regexResult == null) || (regexResult.length != 2))
      return null;

    return regexResult[1];
  }

  this.addParameter = addParameter;
  this.getLastError = getLastError;
  this.getLastErrorCode = getLastErrorCode;
  this.send = send;
  this.stop = stop;
}
